import { AuthenticationClient, RegistrationClient } from "../lib/index.js";

const stepAuth = document.getElementById("step-auth");
const stepRequest = document.getElementById("step-request");
const stepProfile = document.getElementById("step-profile");
const buttonConfirmation = document.getElementById("button-confirmation");
const buttonInstall = document.getElementById("button-install");
const buttonSkip = document.getElementById("button-skip");
const buttonLoginEmail = document.getElementById("button-login-email");
const buttonLoginWebAuthn = document.getElementById("button-login-webauthn");
const buttonAttachWebAuthn = document.getElementById("button-attach-webauthn");

let email = "";
let user = null;

stepAuth.addEventListener("submit", async (event) => {
  event.preventDefault();
  email = event.target.querySelector('[name="email"]').value;
  const { isUser, isWebAuthn } = await checkUser(email);
  if (!email) return;
  buttonConfirmation.hidden = isUser;
  buttonLoginEmail.hidden = !isUser;
  buttonLoginWebAuthn.hidden = !(isUser && isWebAuthn);
});

buttonConfirmation.addEventListener("click", async () => {
  // create user after email confirmation
  await createUser(email);
  // login new user
  user = await authenticateByEmail(email);
  // request biometrics for future login
  toStepRequest();
});

buttonLoginEmail.addEventListener("click", async () => {
  user = await authenticateByEmail(email);
  toStepProfile();
});

buttonInstall.addEventListener("click", async () => {
  const optionsJSON = await webAuthnRegistrationOptions(email);
  const options = RegistrationClient.decodePublicKeyOptionsJSON(optionsJSON);
  let credential;

  try {
    credential = await navigator.credentials.create({ publicKey: options });
  } catch (error) {
    alert(RegistrationClient.getError(error, options));
    return;
  }

  const attestationJSON = RegistrationClient.getAttestationJSON(credential);

  webAuthnRegistration(attestationJSON, email);
  toStepProfile();
});

buttonAttachWebAuthn.addEventListener("click", async () => {
  const optionsJSON = await webAuthnRegistrationOptions(email);
  const options = RegistrationClient.decodePublicKeyOptionsJSON(optionsJSON);
  let credential;

  try {
    credential = await navigator.credentials.create({ publicKey: options });
  } catch (error) {
    alert(RegistrationClient.getError(error, options));
    return;
  }

  const attestationJSON = RegistrationClient.getAttestationJSON(credential);

  webAuthnRegistration(attestationJSON, email);
});

buttonSkip.addEventListener("click", () => {
  toStepProfile();
});

buttonLoginWebAuthn.addEventListener("click", async () => {
  const optionsJSON = await webAuthnAuthenticationOptions(email);
  const options = AuthenticationClient.decodePublicKeyOptionsJSON(optionsJSON);
  let credential;

  try {
    credential = await navigator.credentials.get({ publicKey: options });
  } catch (error) {
    alert(AuthenticationClient.getError(error, options));
    return;
  }

  const assertionJSON = AuthenticationClient.getAssertionJSON(credential);

  try {
    user = await webAuthnAuthentication(assertionJSON, email);
    toStepProfile();
  } catch (error) {
    alert(error);
    return;
  }
});

function toStepRequest() {
  if (!email) throw Error("Email doesn't exist");
  stepAuth.hidden = true;
  stepRequest.hidden = false;
}

function toStepProfile() {
  if (!user) throw Error("User doesn't exist");
  stepAuth.hidden = true;
  stepRequest.hidden = true;
  stepProfile.hidden = false;
  stepProfile.querySelector("h2").textContent = `Greetings, ${user.email}`;
}

async function checkUser(email) {
  const response = await fetch("/user/check", {
    method: "POST",
    body: JSON.stringify({ email }),
  });

  return await response.json();
}

async function authenticateByEmail(email) {
  const response = await fetch(`/email/authentication?email=${email}`);

  return await response.json();
}

async function createUser(email) {
  await fetch("/user", {
    method: "POST",
    body: JSON.stringify({ email }),
  });
}

async function webAuthnRegistrationOptions(email) {
  const response = await fetch("/webauthn/registration", {
    headers: { email },
  });

  return await response.json();
}

async function webAuthnAuthenticationOptions(email) {
  const response = await fetch("/webauthn/authentication", {
    headers: { email },
  });

  return await response.json();
}

async function webAuthnRegistration(payload, email) {
  await fetch("/webauthn/registration", {
    method: "POST",
    headers: { email },
    body: JSON.stringify(payload),
  });
}

async function webAuthnAuthentication(payload, email) {
  const response = await fetch("/webauthn/authentication", {
    method: "POST",
    headers: { email },
    body: JSON.stringify(payload),
  });

  const json = await response.json();

  if (!response.ok) {
    throw Error(json.message);
  }

  return json;
}
