import { Base64URL } from "./helpers/Base64URL.js";

export class AuthenticationClient {
  /**
   * @param {import('./types').Authentication.PublicKeyOptionsJSON} json
   * @returns {PublicKeyCredentialRequestOptions}
   */
  static decodePublicKeyOptionsJSON(json) {
    /** @type {PublicKeyCredentialRequestOptions} */
    const options = {
      challenge: Base64URL.toBuffer(json.challenge),
      rpId: json.rpId,
      userVerification: json.userVerification,
      timeout: json.timeout,
    };

    if (json.allowCredentials) {
      options.allowCredentials = json.allowCredentials.map((credential) => ({
        ...credential,
        id: Base64URL.toBuffer(credential.id),
      }));
    }

    return options;
  }

  /**
   * @param {PublicKeyCredential} credential
   * @returns {import('./types').Authentication.AssertionJSON}
   */
  static getAssertionJSON(credential) {
    const attestationResponse = /** @type {AuthenticatorAssertionResponse} */ (credential.response);

    /** @type {import('./types').Authentication.AssertionJSON} */
    const json = {
      credentialId: credential.id,
      credentialType: credential.type,
      clientDataJSON: Base64URL.fromBuffer(attestationResponse.clientDataJSON),
      authenticatorData: Base64URL.fromBuffer(attestationResponse.authenticatorData),
      signature: Base64URL.fromBuffer(attestationResponse.signature),
    };

    if (attestationResponse.userHandle) {
      json.userHandle = Base64URL.fromBuffer(attestationResponse.userHandle);
    }

    return json;
  }

  /**
   * @param {Error} error
   * @param {PublicKeyCredentialRequestOptions} options
   */
  static getError(error, options) {
    if (error.name === "AbortError") {
      // https://www.w3.org/TR/webauthn-2/#sctn-createCredential (step 16)
      return Error("Registration was sent an abort signal");
    }

    if (error.name === "NotAllowedError") {
      if (options.allowCredentials?.length) {
        // https://www.w3.org/TR/webauthn-2/#sctn-discover-from-external-source (step 17)
        // https://www.w3.org/TR/webauthn-2/#sctn-op-get-assertion (step 6)
        return Error("No available authenticator recognized any of the allowed credentials");
      } else {
        // https://www.w3.org/TR/webauthn-2/#sctn-discover-from-external-source (step 18)
        // https://www.w3.org/TR/webauthn-2/#sctn-op-get-assertion (step 7)
        return Error("User clicked cancel or the authentication process timed out");
      }
    }

    if (error.name === "SecurityError") {
      // https://www.w3.org/TR/webauthn-2/#sctn-discover-from-external-source (step 5)
      // https://www.w3.org/TR/webauthn-2/#sctn-discover-from-external-source (step 6)
      return Error("Invalid domain or RP ID is invalid for this domain");
    }

    if (error.name === "UnknownError") {
      // https://www.w3.org/TR/webauthn-2/#sctn-op-get-assertion (step 1)
      // https://www.w3.org/TR/webauthn-2/#sctn-op-get-assertion (step 12)
      return Error(
        "The authenticator was unable to process provided options or could not create a new assertion signature",
      );
    }

    return Error(error.message);
  }
}
