import { UTF8 } from "./helpers/UTF8.js";
import { SHA256 } from "./helpers/SHA256.js";
import { Base64URL } from "./helpers/Base64URL.js";
import { concatenateBuffers } from "./helpers/concatenateBuffers.js";
import { parseAuthenticatorData } from "./helpers/parseAuthenticatorData.js";

export class AuthenticationServer {
  /**
   * JSON representation of PublicKeyCredentialRequestOptions type
   * https://www.w3.org/TR/webauthn-2/#dictionary-assertion-options
   * @param {import('./types').Authentication.Settings} settings
   * @returns {import('./types').Authentication.PublicKeyOptionsJSON}
   */
  static getPublicKeyOptionsJSON(settings) {
    const challenge = crypto.getRandomValues(new Uint8Array(32));
    /** @type {import('./types').Authentication.PublicKeyOptionsJSON} */
    const options = {
      challenge: Base64URL.fromBuffer(challenge),
      rpId: settings.rpId,
      userVerification: "required",
      timeout: 60000,
    };

    // do not pass empty array
    if (settings.allowCredentialIds) {
      options.allowCredentials = settings.allowCredentialIds.map((id) => ({
        type: "public-key",
        id,
      }));
    }

    return options;
  }

  /**
   * https://www.w3.org/TR/webauthn-2/#sctn-verifying-assertion
   * @param {import('./types').Authentication.AssertionJSON} assertionJSON
   * @param {import('./types').Credential} credential
   * @param {import('./types').Authentication.VerificationOptions} options
   */
  static async verify(assertionJSON, credential, options) {
    const { expectedChallenge, expectedRpId, expectedOrigin } = options;

    if (!assertionJSON.credentialId) {
      throw Error("Attestation must contain credential id");
    }

    if (assertionJSON.credentialType !== "public-key") {
      throw Error("Credential type must be public-key");
    }

    // parse client data
    const clientDataBuffer = Base64URL.toBuffer(assertionJSON.clientDataJSON);
    /** @type {import('./types').ClientDataJSON} */
    const clientDataJSON = JSON.parse(UTF8.fromBuffer(clientDataBuffer));

    // validate type
    if (clientDataJSON.type !== "webauthn.get") {
      throw Error("Response type must be webauthn.get");
    }

    // validate challenge
    if (clientDataJSON.challenge !== expectedChallenge) {
      throw Error("Response challenge does not match the expected");
    }

    // validate origin
    if (!expectedOrigin(clientDataJSON.origin)) {
      throw Error("Response origin does not match the expected");
    }

    // parse authenticatorData
    const authDataBuffer = Base64URL.toBuffer(assertionJSON.authenticatorData);
    const authData = parseAuthenticatorData(authDataBuffer);

    // validate relying part id
    const authDataRpIdHex = SHA256.toHex(authData.rpId);
    const expectedRpIdHash = await SHA256.fromString(expectedRpId);
    const expectedRpIdHex = SHA256.toHex(expectedRpIdHash);
    if (expectedRpIdHex !== authDataRpIdHex) {
      throw Error("Relying party id does not match the expected");
    }

    // expect user presence flag
    if (!authData.flags.up) {
      throw Error("User presence flag expected to be true");
    }

    // expect user verified flag
    if (!authData.flags.uv) {
      throw Error("User verified flag expected to be true");
    }

    // validate counter
    if ((credential.counter > 0 || authData.counter > 0)) {
      if (authData.counter <= credential.counter) {
        throw Error("Invalid counter");
      }
    }

    // validate signature
    const clientDataHash = await SHA256.fromBuffer(clientDataBuffer);
    const concatenatedBuffer = concatenateBuffers(authDataBuffer, clientDataHash);
    const publicKey = Base64URL.toBuffer(credential.publicKey);
    const algorithm = SHA256.identify(credential.algorithm);
    let signature = Base64URL.toBuffer(assertionJSON.signature);

    if (algorithm.name === "ECDSA") {
      signature = SHA256.der2raw(signature);
    }

    const cryptoKey = await crypto.subtle.importKey("spki", publicKey, algorithm, false, ["verify"]);
    const verified = await crypto.subtle.verify(algorithm, cryptoKey, signature, concatenatedBuffer);

    return {
      credentialId: credential.id,
      verified,
      counter: authData.counter,
    };
  }
}
