import { UTF8 } from "./helpers/UTF8.js";
import { Base64URL } from "./helpers/Base64URL.js";

export class RegistrationClient {
  /**
   * @param {import('./types').Registration.PublicKeyOptionsJSON} json
   * @returns {PublicKeyCredentialCreationOptions}
   */
  static decodePublicKeyOptionsJSON(json) {
    /** @type {PublicKeyCredentialCreationOptions} */
    const options = {
      challenge: Base64URL.toBuffer(json.challenge),
      rp: json.rp,
      user: {
        id: UTF8.toBuffer(json.user.id),
        name: json.user.name,
        displayName: json.user.displayName,
      },
      pubKeyCredParams: json.pubKeyCredParams,
      authenticatorSelection: json.authenticatorSelection,
      attestation: json.attestation,
      timeout: json.timeout,
    };

    if (json.excludeCredentials) {
      options.excludeCredentials = json.excludeCredentials.map((credential) => ({
        ...credential,
        id: Base64URL.toBuffer(credential.id),
      }));
    }

    return options;
  }

  /**
   * @param {PublicKeyCredential} credential
   * @returns {import('./types').Registration.AttestationJSON}
   */
  static getAttestationJSON(credential) {
    const attestationResponse = /** @type {AuthenticatorAttestationResponse} */ (credential.response);
    const publicKey = attestationResponse.getPublicKey();

    if (!publicKey) throw Error("Failed to get public key from attestation response");

    return {
      credentialId: credential.id,
      credentialType: credential.type,
      clientDataJSON: Base64URL.fromBuffer(attestationResponse.clientDataJSON),
      publicKey: Base64URL.fromBuffer(publicKey),
      algorithm: attestationResponse.getPublicKeyAlgorithm(),
      authenticatorData: Base64URL.fromBuffer(attestationResponse.getAuthenticatorData()),
    };
  }

  /**
   * @param {Error} error
   * @param {PublicKeyCredentialCreationOptions} options
   */
  static getError(error, options) {
    if (error.name === "NotSupportedError") {
      // https://www.w3.org/TR/webauthn-2/#sctn-op-make-cred (step 2, step 10)
      return Error("No available authenticator supported any of the specified algorithms");
    }

    if (error.name === "NotAllowedError") {
      // https://www.w3.org/TR/webauthn-2/#sctn-createCredential (step 20, step 21)
      return Error("User clicked cancel or the registration timed out");
    }

    if (error.name === "AbortError") {
      // https://www.w3.org/TR/webauthn-2/#sctn-createCredential (step 16)
      return Error("Registration was sent an abort signal");
    }

    if (error.name === "ConstraintError") {
      if (options.authenticatorSelection?.residentKey === "required") {
        // https://www.w3.org/TR/webauthn-2/#sctn-op-make-cred (step 4)
        return Error("Discoverable credentials were required but no available authenticator supported it");
      } else if (options.authenticatorSelection?.userVerification === "required") {
        // https://www.w3.org/TR/webauthn-2/#sctn-op-make-cred (step 5)
        return Error("User verification was required but no available authenticator supported it");
      }
    }

    if (error.name === "InvalidStateError") {
      // https://www.w3.org/TR/webauthn-2/#sctn-createCredential (step 20)
      // https://www.w3.org/TR/webauthn-2/#sctn-op-make-cred (step 3)
      return Error("The authenticator was previously registered");
    }

    if (error.name === "SecurityError") {
      // https://www.w3.org/TR/webauthn-2/#sctn-createCredential (step 7)
      // https://www.w3.org/TR/webauthn-2/#sctn-createCredential (step 8)
      return Error("Invalid domain or RP ID is invalid for this domain");
    }

    if (error.name === "TypeError") {
      if (options.user.id.byteLength < 1 || options.user.id.byteLength > 64) {
        // https://www.w3.org/TR/webauthn-2/#sctn-createCredential (step 5)
        return Error("User ID was not between 1 and 64 characters");
      }
    }

    if (error.name === "UnknownError") {
      // https://www.w3.org/TR/webauthn-2/#sctn-op-make-cred (step 1, step 8)
      return Error("The authenticator was unable to process the specified options or could not create a new credential");
    }

    return Error(error.message);
  }
}
