import { UTF8 } from "./helpers/UTF8.js";
import { SHA256 } from "./helpers/SHA256.js";
import { Base64URL } from "./helpers/Base64URL.js";
import { parseAuthenticatorData } from "./helpers/parseAuthenticatorData.js";

export class RegistrationServer {
  /**
   * JSON representation of PublicKeyCredentialCreationOptions type
   * https://www.w3.org/TR/webauthn-2/#dictionary-makecredentialoptions
   * @param {import('./types').Registration.Settings} settings
   * @returns {import('./types').Registration.PublicKeyOptionsJSON}
   */
  static getPublicKeyOptionsJSON(settings) {
    const challenge = crypto.getRandomValues(new Uint8Array(32));
    /** @type {import('./types').Registration.PublicKeyOptionsJSON} */
    const options = {
      challenge: Base64URL.fromBuffer(challenge),
      rp: {
        id: settings.rpId,
        name: settings.rpName,
      },
      user: {
        id: settings.userId,
        name: settings.userName,
        displayName: settings.userName,
      },
      pubKeyCredParams: [
        { type: "public-key", alg: -7 }, // ES256
        { type: "public-key", alg: -257 }, // RS256
      ],
      authenticatorSelection: {
        userVerification: "required",
        authenticatorAttachment: "platform",
      },
      attestation: "none",
      timeout: 60000,
    };

    if (settings.excludeCredentialIds) {
      options.excludeCredentials = settings.excludeCredentialIds.map((id) => ({
        type: "public-key",
        id,
      }));
    }

    return options;
  }

  /**
   * https://www.w3.org/TR/webauthn-2/#sctn-registering-a-new-credential
   * @param {import('./types').Registration.AttestationJSON} attestationJSON
   * @param {import('./types').Registration.VerificationOptions} options
   * @returns {Promise<import('./types').Credential>}
   */
  static async validate(attestationJSON, options) {
    const { expectedChallenge, expectedRpId, expectedOrigin } = options;

    if (!attestationJSON.credentialId) {
      throw Error("Attestation must contain credential id");
    }

    if (attestationJSON.credentialType !== "public-key") {
      throw Error("Credential type must be public-key");
    }

    // parse client data
    const clientDataBuffer = Base64URL.toBuffer(attestationJSON.clientDataJSON);
    /** @type {import('./types').ClientDataJSON} */
    const clientDataJSON = JSON.parse(UTF8.fromBuffer(clientDataBuffer));

    // validate type
    if (clientDataJSON.type !== "webauthn.create") {
      throw Error("Response type must be webauthn.create");
    }

    // validate challenge
    if (clientDataJSON.challenge !== expectedChallenge) {
      throw Error("Response challenge does not match the expected");
    }

    // validate origin
    if (!expectedOrigin(clientDataJSON.origin)) {
      throw Error("Response origin does not match the expected");
    }

    // algorithm should be one of those passed to publicKey creation options
    if (![-7, -256].includes(attestationJSON.algorithm)) {
      throw Error("Algorithm code must be one of: -7, -257");
    }

    // parse authenticatorData
    const authDataBuffer = Base64URL.toBuffer(attestationJSON.authenticatorData);
    const authData = parseAuthenticatorData(authDataBuffer);

    // validate relying part id
    const authDataRpIdHex = SHA256.toHex(authData.rpId);
    const expectedRpIdHash = await SHA256.fromString(expectedRpId);
    const expectedRpIdHex = SHA256.toHex(expectedRpIdHash);
    if (expectedRpIdHex !== authDataRpIdHex) {
      throw Error("Relying party id does not match the expected");
    }

    // expect user presence flag
    if (!authData.flags.up) {
      throw Error("User presence flag expected to be true");
    }

    // expect user verified flag
    if (!authData.flags.uv) {
      throw Error("User verified flag expected to be true");
    }

    return {
      id: attestationJSON.credentialId,
      publicKey: attestationJSON.publicKey,
      algorithm: attestationJSON.algorithm,
      counter: authData.counter,
    };
  }
}
