import { Database } from "https://deno.land/x/sqlite3@0.6.1/mod.ts";

const db = new Database("database.db");

export interface User {
  id: string;
  email: string;
  challenge: string;
}

export interface Credential {
  id: string;
  userId: string;
  publicKey: string;
  algorithm: number;
  counter: number;
}

db.exec(`
  CREATE TABLE IF NOT EXISTS user (
    id TEXT PRIMARY KEY,
    email TEXT NOT NULL UNIQUE,
    challenge TEXT
  )
`);

db.exec(`
  CREATE TABLE IF NOT EXISTS credential (
    id TEXT PRIMARY KEY,
    userId TEXT NOT NULL,
    publicKey TEXT NOT NULL,
    algorithm TEXT NOT NULL,
    counter INTEGER NOT NULL
  )
`);

function insertUser(email: string) {
  const id = crypto.randomUUID();
  db.exec(
    "INSERT INTO user (id, email) VALUES(?, ?)",
    id,
    email,
  );
}

function updateChallenge(id: string, challenge: string) {
  db.exec(
    "UPDATE user SET challenge = ? WHERE id = ?",
    challenge,
    id,
  );
}

function selectUser(user: Partial<User>): User | null {
  let record;
  if (user.id) {
    record = db
      .prepare("SELECT id, email, challenge FROM user WHERE id = ?")
      .get(user.id) as unknown as User;
  }
  if (user.email) {
    record = db
      .prepare("SELECT id, email, challenge FROM user WHERE email = ?")
      .get(user.email) as unknown as User;
  }
  return record ?? null;
}

function insertCredential(credential: Credential) {
  db.exec(
    "INSERT INTO credential (id, userId, publicKey, algorithm, counter) VALUES(?, ?, ?, ?, ?)",
    credential.id,
    credential.userId,
    credential.publicKey,
    credential.algorithm,
    credential.counter,
  );
}

function updateCounter(credentialId: string, counter: number) {
  db.exec(
    "UPDATE credential SET counter = ? WHERE id = ?",
    counter,
    credentialId,
  );
}

function selectCredential(credentialId: string, userId: string): Credential {
  return db
    .prepare("SELECT * FROM credential WHERE id = ? AND userId = ?")
    .get(credentialId, userId) as unknown as Credential;
}

function selectCredentials(userId: string): Credential[] {
  return db
    .prepare("SELECT * FROM credential WHERE userId = ?")
    .all(userId) as unknown as Credential[];
}

export default {
  insertUser,
  updateChallenge,
  selectUser,
  insertCredential,
  selectCredential,
  selectCredentials,
  updateCounter,
};
