import db from "./database.ts";
import { json_error, json_success } from "./json_response.ts";
import { AuthenticationServer, RegistrationServer } from "../lib/index.js";
import config from "./config.ts";

async function user(request: Request): Promise<Response> {
  if (request.method === "POST") {
    const json: { email: string } = await request.json();
    db.insertUser(json.email);

    return json_success(null, 201);
  }

  throw Error("Unknown HTTP method");
}

async function checkUser(request: Request): Promise<Response> {
  const json: { email: string } = await request.json();
  const user = db.selectUser({ email: json.email });
  let isWebAuthn = false;
  if (user && db.selectCredentials(user.id).length) {
    isWebAuthn = true;
  }

  return json_success({ isUser: Boolean(user), isWebAuthn });
}

function emailAuthentication(request: Request): Promise<Response> {
  // in real project user email should not be passed in open form (e.g. pass expirable UUID through email link instead)
  const url = new URL(request.url);
  const email = url.searchParams.get("email");

  if (!email) return Promise.resolve(json_error("Email is required", 400));
  const user = db.selectUser({ email });
  if (!user) return Promise.resolve(json_error("Bad credentials", 400));

  return Promise.resolve(json_success({
    id: user.id,
    email: user.email,
  }));
}

async function webAuthnRegistration(request: Request): Promise<Response> {
  // get user
  const email = request.headers.get("email");
  if (!email) return Promise.resolve(json_error("Email is required", 400));
  const user = db.selectUser({ email });
  if (!user) return Promise.resolve(json_error("Bad credentials", 400));

  // return options for GET request
  if (request.method === "GET") {
    // exclude authenticator that has one of these registered already credentials
    const existingCredentialIds = db.selectCredentials(user.id).map((c) => c.id);

    const options = RegistrationServer.getPublicKeyOptionsJSON({
      rpId: config.rp.id,
      rpName: config.rp.name,
      userId: user.id,
      userName: user.email,
      excludeCredentialIds: existingCredentialIds,
    });

    // save new challenge for attestation verification
    db.updateChallenge(user.id, options.challenge);

    return Promise.resolve(json_success(options));
  }

  // do validation on POST request
  if (request.method === "POST") {
    const attestationJSON = await request.json();

    try {
      const credential = await RegistrationServer.validate(attestationJSON, {
        expectedRpId: config.rp.id,
        expectedChallenge: user.challenge,
        expectedOrigin: (origin: string) => new URL(origin).hostname === "localhost",
      });

      db.insertCredential({
        id: credential.id,
        userId: user.id,
        publicKey: credential.publicKey,
        algorithm: credential.algorithm,
        counter: credential.counter,
      });
    } catch (error) {
      return Promise.resolve(json_error(error.message, 400));
    }

    return Promise.resolve(json_success(null));
  }

  throw Error("Unknown HTTP method");
}

async function webAuthnAuthentication(request: Request): Promise<Response> {
  // get user
  const email = request.headers.get("email");
  if (!email) return Promise.resolve(json_error("Email is required", 400));
  const user = db.selectUser({ email });
  if (!user) return Promise.resolve(json_error("Bad credentials", 400));

  // return options for GET request
  if (request.method === "GET") {
    const credentials = db.selectCredentials(user.id);
    const options = AuthenticationServer.getPublicKeyOptionsJSON({
      rpId: config.rp.id,
      allowCredentialIds: credentials.map((credential) => credential.id),
    });

    // save new challenge for assertion verification
    db.updateChallenge(user.id, options.challenge);

    return Promise.resolve(json_success(options));
  }

  // do validation on POST request
  if (request.method === "POST") {
    const assertionJSON = await request.json();
    const { credentialId } = assertionJSON;

    const credential = db.selectCredential(credentialId, user.id);
    if (!credential) {
      return json_error("Invalid credential id", 400);
    }

    const { verified, counter } = await AuthenticationServer.verify(assertionJSON, credential, {
      expectedChallenge: user.challenge,
      expectedRpId: config.rp.id,
      expectedOrigin: (origin: string) => new URL(origin).hostname === "localhost",
    });

    if (verified) {
      db.updateCounter(credentialId, counter);

      // return user for successful authentication
      return json_success({
        id: user.id,
        email: user.email,
      });
    }

    return json_error("Validation failed", 403);
  }

  throw Error("Unknown HTTP method");
}

export default {
  user,
  checkUser,
  emailAuthentication,
  webAuthnRegistration,
  webAuthnAuthentication,
};
